from iohub.ngff.models import TransformationMeta
from iohub.ngff.nodes import (
    ImageArray,
    NGFFNode,
    Plate,
    Position,
    TiledImageArray,
    TiledPosition,
    open_ome_zarr,
)

__all__ = [
    "ImageArray",
    "NGFFNode",
    "Plate",
    "Position",
    "TiledImageArray",
    "TiledPosition",
    "TransformationMeta",
    "open_ome_zarr",
]



from __future__ import annotations

import logging
from copy import copy
from pathlib import Path
from typing import TYPE_CHECKING, Iterable
from warnings import catch_warnings, filterwarnings

import dask.array as da
import numpy as np
import zarr
from natsort import natsorted
from numpy.typing import ArrayLike
from tifffile import TiffFile
from xarray import DataArray

from iohub.mm_fov import MicroManagerFOV, MicroManagerFOVMapping

if TYPE_CHECKING:
    from _typeshed import StrOrBytesPath


__all__ = ["MMOmeTiffFOV", "MMStack"]
_logger = logging.getLogger(__name__)


def _normalize_mm_pos_key(key: str | int) -> int:
    try:
        return int(key)
    except TypeError:
        raise TypeError("Micro-Manager position keys must be integers.")


def find_first_ome_tiff_in_mmstack(data_path: Path) -> Path:
    if data_path.is_file():
        if "ome.tif" in data_path.name:
            return data_path
        else:
            raise ValueError("{data_path} is not a OME-TIFF file.")
    elif data_path.is_dir():
        files = data_path.glob("*.ome.tif")
        try:
            return next(files)
        except StopIteration:
            raise FileNotFoundError(
                f"Path {data_path} contains no OME-TIFF files."
            )
    raise FileNotFoundError(f"Path {data_path} does not exist.")


class MMOmeTiffFOV(MicroManagerFOV):
    def __init__(self, parent: MMStack, key: str) -> None:
        super().__init__(parent, key)
        self._xdata = parent.xdata[key]

    @property
    def axes_names(self) -> list[str]:
        return list(self._xdata.dims)

    @property
    def shape(self) -> tuple[int, int, int, int, int]:
        return self._xdata.shape

    @property
    def dtype(self) -> np.dtype:
        return self._xdata.dtype

    @property
    def t_scale(self) -> float:
        return self.parent._t_scale

    def __getitem__(self, key: int | slice | tuple[int | slice]) -> ArrayLike:
        return self._xdata[key]

    @property
    def xdata(self) -> DataArray:
        return self._xdata

    def frame_metadata(self, t: int, c: int, z: int) -> dict | None:
        """Read image plane metadata from the OME-TIFF file."""
        return self.parent.read_image_metadata(self._position, t, c, z)


class MMStack(MicroManagerFOVMapping):
    """Micro-Manager multi-file OME-TIFF (MMStack) reader.

    Parameters
    ----------
    data_path : StrOrBytesPath
        Path to the directory containing OME-TIFF files
        or the path to the first OME-TIFF file in the series
    """

    def __init__(self, data_path: StrOrBytesPath):
        super().__init__()
        data_path = Path(data_path)
        first_file = find_first_ome_tiff_in_mmstack(data_path)
        self._root = first_file.parent
        self.dirname = self._root.name
        self._first_tif = TiffFile(first_file, is_mmstack=True)
        _logger.debug(f"Parsing {first_file} as MMStack.")
        with catch_warnings():
            # The IJMetadata tag (50839) is sometimes not written
            # See https://micro-manager.org/Micro-Manager_File_Formats
            filterwarnings("ignore", message=r".*50839.*", module="tifffile")
            self._parse_data()
        self._store = None

    def _parse_data(self):
        series = self._first_tif.series[0]
        raw_dims = dict(
            (axis, size)
            for axis, size in zip(series.get_axes(), series.get_shape())
        )
        axes = ("R", "T", "C", "Z", "Y", "X")
        dims = dict((ax, raw_dims.get(ax, 1)) for ax in axes)
        _logger.debug(f"Got dataset dimensions from tifffile: {dims}.")
        (
            self.positions,
            self.frames,
            self.channels,
            self.slices,
            self.height,
            self.width,
        ) = dims.values()
        self._set_mm_meta(self._first_tif.micromanager_metadata)
        self._store = series.aszarr()
        _logger.debug(f"Opened {self._store}.")
        data = da.from_zarr(zarr.open(self._store))
        self.dtype = data.dtype
        img = DataArray(data, dims=raw_dims, name=self.dirname)
        xarr = img.expand_dims(
            [ax for ax in axes if ax not in img.dims]
        ).transpose(*axes)
        if self.channels > len(self.channel_names):
            for c in range(self.channels):
                if c >= len(self.channel_names):
                    self.channel_names.append(f"Channel{c}")
            _logger.warning(
                "Number of channel names in the metadata is "
                "smaller than the number of channels. "
                f"Completing with fallback names: {self.channel_names}."
            )
        # not all positions in the position list may have been acquired
        xarr = xarr[: self.positions]
        xarr.coords["C"] = self.channel_names
        xset = xarr.to_dataset(dim="R")
        self._xdata = xset
        self._infer_image_meta()

    @property
    def xdata(self):
        return self._xdata

    def __len__(self) -> int:
        return len(self.xdata.keys())

    def __getitem__(self, key: str | int) -> MMOmeTiffFOV:
        key = _normalize_mm_pos_key(key)
        return MMOmeTiffFOV(self, key)

    def __setitem__(self, key, value) -> None:
        raise PermissionError("MMStack is read-only.")

    def __delitem__(self, key, value) -> None:
        raise PermissionError("MMStack is read-only.")

    def __contains__(self, key: str | int) -> bool:
        key = _normalize_mm_pos_key(key)
        return key in self.xdata

    def __iter__(self) -> Iterable[tuple[str, MMOmeTiffFOV]]:
        for key in self.xdata:
            yield str(key), self[key]

    def __enter__(self) -> MMStack:
        return self

    def __exit__(self, exc_type, exc_val, exc_tb) -> None:
        self.close()

    def close(self) -> None:
        """Close file handles"""
        self._first_tif.close()

    def _set_mm_meta(self, mm_meta: dict) -> None:
        """Assign image metadata from summary metadata."""
        self._mm_meta = mm_meta
        self.channel_names = []
        mm_version = self._mm_meta["Summary"]["MicroManagerVersion"]
        if "beta" in mm_version:
            if self._mm_meta["Summary"]["Positions"] > 1:
                self._stage_positions = []

                for p in range(
                    len(self._mm_meta["Summary"]["StagePositions"])
                ):
                    pos = self._simplify_stage_position_beta(
                        self._mm_meta["Summary"]["StagePositions"][p]
                    )
                    self._stage_positions.append(pos)

            # MM beta versions sometimes don't have 'ChNames',
            # so I'm wrapping in a try-except and setting the
            # channel names to empty strings if it fails.
            try:
                for ch in self._mm_meta["Summary"]["ChNames"]:
                    self.channel_names.append(ch)
            except Exception:
                self.channel_names = self._mm_meta["Summary"]["Channels"] * [
                    ""
                ]  # empty strings

        elif mm_version == "1.4.22":
            for ch in self._mm_meta["Summary"].get("ChNames", []):
                self.channel_names.append(ch)

        # Parsing of data acquired with the OpenCell
        # acquisition script on the Dragonfly miroscope
        elif (
            mm_version == "2.0.1 20220920"
            and self._mm_meta["Summary"].get("Prefix", None) == "raw_data"
        ):
            files = natsorted(self.root.glob("*.ome.tif"))
            self.positions = len(files)  # not all positions are saved

            if self._mm_meta["Summary"]["Positions"] > 1:
                self._stage_positions = [None] * self.positions

                for p_idx, file_name in enumerate(files):
                    site_idx = int(str(file_name).split("_")[-1].split("-")[0])
                    pos = self._simplify_stage_position(
                        self._mm_meta["Summary"]["StagePositions"][site_idx]
                    )
                    self._stage_positions[p_idx] = pos

            for ch in self._mm_meta["Summary"]["ChNames"]:
                self.channel_names.append(ch)

        else:
            if self._mm_meta["Summary"].get("Positions", 1) > 1:
                self._stage_positions = []

                for p in range(self._mm_meta["Summary"]["Positions"]):
                    pos = self._simplify_stage_position(
                        self._mm_meta["Summary"]["StagePositions"][p]
                    )
                    self._stage_positions.append(pos)

            for ch in self._mm_meta["Summary"].get("ChNames", []):
                self.channel_names.append(ch)
        z_step_size = float(self._mm_meta["Summary"].get("z-step_um", 1.0))
        if z_step_size == 0:
            if self.slices == 1:
                z_step_size = 1.0
            else:
                _logger.warning(
                    f"Z-step size is {z_step_size} um in the metadata, "
                    "Using 1.0 um instead."
                )
        self._z_step_size = z_step_size
        self.height = self._mm_meta["Summary"]["Height"]
        self.width = self._mm_meta["Summary"]["Width"]
        self._t_scale = (
            float(self._mm_meta["Summary"].get("Interval_ms", 1e3)) / 1e3
        )

    def _simplify_stage_position(self, stage_pos: dict):
        """
        flattens the nested dictionary structure of stage_pos
        and removes superfluous keys

        Parameters
        ----------
        stage_pos:      (dict)
            dictionary containing a single position's device info

        Returns
        -------
        out:            (dict)
            flattened dictionary
        """

        out = copy(stage_pos)
        out.pop("DevicePositions")
        for dev_pos in stage_pos["DevicePositions"]:
            out.update({dev_pos["Device"]: dev_pos["Position_um"]})
        return out

    def _simplify_stage_position_beta(self, stage_pos: dict) -> dict:
        """
        flattens the nested dictionary structure of stage_pos
        and removes superfluous keys
        for MM2.0 Beta versions

        Parameters
        ----------
        stage_pos:      (dict)
            dictionary containing a single position's device info

        Returns
        -------
        new_dict:       (dict)
            flattened dictionary

        """

        new_dict = {}
        new_dict["Label"] = stage_pos["label"]
        new_dict["GridRow"] = stage_pos["gridRow"]
        new_dict["GridCol"] = stage_pos["gridCol"]

        for sub in stage_pos["subpositions"]:
            values = []
            for field in ["x", "y", "z"]:
                if sub[field] != 0:
                    values.append(sub[field])
            if len(values) == 1:
                new_dict[sub["stageName"]] = values[0]
            else:
                new_dict[sub["stageName"]] = values

        return new_dict

    def read_image_metadata(
        self, p: int, t: int, c: int, z: int
    ) -> dict | None:
        """Read image plane metadata from the OME-TIFF file."""
        multi_index = (p, t, c, z)
        tif_shape = (self.positions, self.frames, self.channels, self.slices)
        idx = np.ravel_multi_index(multi_index, tif_shape)
        return self._read_frame_metadata(idx)

    def _read_frame_metadata(self, idx: int) -> dict | None:
        # `TiffPageSeries` is not a collection of `TiffPage` objects
        # but a mixture of `TiffPage` and `TiffFrame` objects
        # https://github.com/cgohlke/tifffile/issues/179
        with catch_warnings():
            filterwarnings(
                "ignore", message=r".*from closed file.*", module="tifffile"
            )
            try:
                # virtual frames
                page = self._first_tif.pages[idx]
            except IndexError:
                page = self._first_tif.series[0].pages[idx]
            if page:
                try:
                    page = page.aspage()
                except ValueError:
                    _logger.warning("Cannot read tags from virtual frame.")
                    return None
            else:
                # invalid page
                _logger.warning(f"Page {idx} is not present in the dataset.")
                return None
            try:
                return page.tags["MicroManagerMetadata"].value
            except KeyError:
                _logger.warning("The Micro-Manager metadata tag is not found.")
                return None

    def _infer_image_meta(self) -> None:
        """
        Infer data type and pixel size from the first image plane metadata.
        """
        _logger.debug("Inferring image metadata.")
        metadata = self._read_frame_metadata(0)
        if metadata is not None:
            try:
                self._xy_pixel_size = float(metadata["PixelSizeUm"])
                if self._xy_pixel_size > 0:
                    return
            except Exception:
                _logger.warning(
                    "Micro-Manager image plane metadata cannot be loaded."
                )
        _logger.warning(
            "XY pixel size cannot be determined, defaulting to 1.0 um."
        )
        self._xy_pixel_size = 1.0

    @property
    def zyx_scale(self) -> tuple[float, float, float]:
        return (
            self._z_step_size,
            self._xy_pixel_size,
            self._xy_pixel_size,
        )



import numpy as np
import zarr

from iohub._deprecated.zarrfile import ZarrReader
from iohub.reader import read_images
from tests.conftest import mm2gamma_zarr_v01


def test_constructor_mm2gamma():
    """
    test that constructor parses metadata properly
        no data extraction in this test
    """
    mmr = read_images(mm2gamma_zarr_v01)
    assert isinstance(mmr, ZarrReader)
    assert mmr.micromanager_metadata is not None
    assert mmr.z_step_size is not None
    assert mmr.width > 0
    assert mmr.height > 0
    assert mmr.frames > 0
    assert mmr.slices > 0
    assert mmr.channels > 0
    assert mmr.rows is not None
    assert mmr.columns is not None
    assert mmr.wells is not None
    assert mmr.hcs_meta is not None
    # Check HCS metadata copy
    meta = mmr.hcs_meta
    assert "plate" in meta.keys()
    assert "well" in meta.keys()
    assert len(meta["well"]) == mmr.get_num_positions()
    assert "images" in meta["well"][0]
    assert len(meta["well"][0]["images"]) != 0
    assert "path" in meta["well"][0]["images"][0]
    assert meta["well"][0]["images"][0]["path"] == "Pos_000"


def test_output_dims_mm2gamma():
    """
    test that output dimensions are always (t, c, z, y, x)
    """
    mmr = ZarrReader(mm2gamma_zarr_v01)
    assert mmr.get_array(0).shape[0] == mmr.frames
    assert mmr.get_array(0).shape[1] == mmr.channels
    assert mmr.get_array(0).shape[2] == mmr.slices
    assert mmr.get_array(0).shape[3] == mmr.height
    assert mmr.get_array(0).shape[4] == mmr.width


def test_get_zarr_mm2gamma():
    mmr = ZarrReader(mm2gamma_zarr_v01)
    for i in range(mmr.get_num_positions()):
        z = mmr.get_zarr(i)
        assert isinstance(z, zarr.core.Array)


def test_get_array_mm2gamma():
    mmr = ZarrReader(mm2gamma_zarr_v01)
    for i in range(mmr.get_num_positions()):
        z = mmr.get_array(i)
        assert z.shape == mmr.shape
        assert isinstance(z, np.ndarray)


def test_get_image_mm2gamma():
    mmr = ZarrReader(mm2gamma_zarr_v01)
    for i in range(mmr.get_num_positions()):
        z = mmr.get_image(i, t=0, c=0, z=0)
        assert z.shape == (mmr.shape[-2], mmr.shape[-1])
        assert isinstance(z, np.ndarray)


def test_get_num_positions_mm2gamma():
    mmr = ZarrReader(mm2gamma_zarr_v01)
    assert mmr.get_num_positions() == 4



import re

import pytest
from xarray import DataArray

from iohub.ndtiff import NDTiffDataset, NDTiffFOV
from tests.conftest import ndtiff_v2_datasets, ndtiff_v3_labeled_positions


def pytest_generate_tests(metafunc):
    if "ndtiff_dataset" in metafunc.fixturenames:
        metafunc.parametrize(
            "ndtiff_dataset",
            ndtiff_v2_datasets + [ndtiff_v3_labeled_positions],
        )


def test_dataset_ctx(ndtiff_dataset):
    with NDTiffDataset(ndtiff_dataset) as dataset:
        assert isinstance(dataset, NDTiffDataset)
        assert len(dataset) > 0
        assert "NDTiffDataset" in dataset.__repr__()


def test_dataset_nonexisting(tmpdir):
    with pytest.raises(FileNotFoundError):
        NDTiffDataset(tmpdir / "nonexisting")


def test_dataset_metadata(ndtiff_dataset):
    with NDTiffDataset(ndtiff_dataset) as dataset:
        assert isinstance(dataset.micromanager_metadata, dict)
        assert dataset.micromanager_metadata["Summary"]
        assert isinstance(dataset.micromanager_summary, dict)
        if dataset.stage_positions:
            assert "DefaultXYStage" in dataset.stage_positions[0]


@pytest.mark.parametrize("ndtiff_v2", ndtiff_v2_datasets)
def test_dataset_getitem_v2(ndtiff_v2):
    with NDTiffDataset(ndtiff_v2) as dataset:
        assert isinstance(dataset["0"], NDTiffFOV)
        assert isinstance(dataset[0], NDTiffFOV)


def test_dataset_v3_labeled_positions():
    dataset = NDTiffDataset(ndtiff_v3_labeled_positions)
    assert len(dataset) == 3
    positions = ["Pos0", "Pos1", "Pos2"]
    for (key, fov), name in zip(dataset, positions):
        assert key == name
        assert isinstance(fov, NDTiffFOV)
        assert name in dataset.__repr__()
        assert key in fov.__repr__()
    with pytest.raises(KeyError):
        dataset["0"]
        dataset[0]
    dataset.close()


def test_dataset_iter(ndtiff_dataset):
    with NDTiffDataset(ndtiff_dataset) as dataset:
        for key, fov in dataset:
            assert isinstance(key, str)
            assert isinstance(fov, NDTiffFOV)


def test_dataset_v2_num_positions(ndtiff_dataset):
    with NDTiffDataset(ndtiff_dataset) as dataset:
        p_match = re.search(r"_(\d+)p_", str(ndtiff_dataset))
        assert len(dataset) == int(p_match.group(1)) if p_match else 1


def test_fov_getitem(ndtiff_dataset):
    with NDTiffDataset(ndtiff_dataset) as dataset:
        for _, fov in dataset:
            img = fov[:]
            assert isinstance(img, DataArray)
            assert img.ndim == 5
            assert img[0, 0, 0, 0, 0] >= 0
            for ch in fov.channel_names:
                assert img.sel(time=0, channel=ch, z=0, y=0, x=0) >= 0



from pathlib import Path
from typing import Any

import pytest

from iohub.fov import BaseFOV, FOVDict


class FOV(BaseFOV):
    def __init__(self, axes: list[str]) -> None:
        super().__init__()
        self._axes = axes

    @property
    def root(self) -> Path:
        return Path()

    @property
    def axes_names(self) -> list[str]:
        return self._axes

    @property
    def channel_names(self) -> list[str]:
        return []

    def __getitem__(self, key: Any) -> Any:
        pass

    @property
    def dtype(self) -> Any:
        pass

    @property
    def shape(self) -> Any:
        pass

    @property
    def zyx_scale(self) -> tuple[float, float, float]:
        return (1.0,) * 3

    @property
    def t_scale(self) -> float:
        raise 1.0


@pytest.mark.parametrize(
    "axes,missing",
    [
        (["T", "C", "Z", "Y", "X"], []),
        (["time", "z", "y", "x"], [1]),
        (["channels", "Y", "X"], [0, 2]),
    ],
)
def test_missing_axes(axes: list[str], missing: list[int]) -> None:
    fov = FOV(axes)
    assert fov._missing_axes() == missing

    shape = (10,) * len(axes)
    padded_shape = fov._pad_missing_axes(shape, 1)
    assert len(padded_shape) == 5

    for i, s in enumerate(padded_shape):
        if i in missing:
            assert s == 1
        else:
            assert s == 10


def test_fov_dict() -> None:

    good_collection = FOVDict(
        {
            "488": FOV(["y", "x"]),
            "561": FOV(["y", "x"]),
        },
        mask=FOV(["c", "x", "y"]),
    )

    assert len(good_collection) == 3
    assert "488" in good_collection
    assert good_collection["mask"] is not None

    with pytest.raises(TypeError):
        del good_collection["561"]

    with pytest.raises(TypeError):
        good_collection["segmentation"] = FOV(["x", "y"])

    with pytest.raises(TypeError):
        FOVDict({488: FOV(["y", "x"])})

    with pytest.raises(TypeError):
        FOVDict(mask=[1, 2, 3])



import os
import csv
import shutil
from pathlib import Path

import fsspec
import pytest
from wget import download


def _download_ndtiff_v3_labeled_positions(test_data: Path) -> None:
    ghfs = fsspec.filesystem(
        "github",
        org="micro-manager",
        repo="NDTiffStorage",
        username=os.environ.get("GITHUB_ACTOR"),
        token=os.environ.get("GITHUB_TOKEN"),
    )
    v3_lp = test_data / "ndtiff_v3_labeled_positions"
    Path.mkdir(v3_lp)
    ghfs.get(
        ghfs.ls("test_data/v3/labeled_positions_1"),
        str(v3_lp),
        recursive=True,
    )


def download_data():
    """Download test datasets."""
    test_data = Path.cwd() / ".pytest_temp" / "test_data"
    if not test_data.is_dir():
        Path.mkdir(test_data, parents=True)
        print("\nsetting up temp folder")

    # Zenodo URL
    custom_url = (
        "https://zenodo.org/record/6983916/files/waveOrder_test_data.zip"
    )
    # Reference v0.4 HCS dataset from OME
    # See the last line of
    # https://github.com/ome/ngff/issues/140#issuecomment-1309972511
    ome_hcs_url = "https://zenodo.org/record/8091756/files/20200812-CardiomyocyteDifferentiation14-Cycle1.zarr.zip"  # noqa

    # download files to temp folder
    if not any(test_data.iterdir()):
        print("Downloading test files...")
        for url in (custom_url, ome_hcs_url):
            output = test_data / Path(url).name
            download(url, out=str(output))
            shutil.unpack_archive(output, extract_dir=test_data)
        _download_ndtiff_v3_labeled_positions(test_data)
    return test_data


def subdirs(parent: Path, name: str) -> list[Path]:
    return [d for d in (parent / name).iterdir() if d.is_dir()]


test_datasets = download_data()


mm2gamma_ome_tiffs = subdirs(test_datasets, "MM20_ome-tiffs")


mm2gamma_ome_tiffs_hcs = [p for p in mm2gamma_ome_tiffs if "4p" in p.name]


# This is a dataset with 11 timepoints
# The MDA definition at start of the experiment specifies 20 timepoints
mm2gamma_ome_tiffs_incomplete = (
    test_datasets
    / "MM20_ometiff_incomplete"
    / "mm2.0-20201209_20t_5z_3c_512k_incomplete_1"
)


mm2gamma_singlepage_tiffs = subdirs(test_datasets, "MM20_singlepage-tiffs")


# This is a dataset with 11 timepoints
# The MDA definition at start of the experiment specifies 20 timepoints
mm2gamma_singlepage_tiffs_incomplete = (
    test_datasets
    / "MM20_singlepage_incomplete"
    / "mm2.0-20201209_20t_5z_3c_512k_incomplete_1 2"
)


mm1422_ome_tiffs = subdirs(test_datasets, "MM1422_ome-tiffs")


mm1422_singlepage_tiffs = subdirs(test_datasets, "MM1422_singlepage-tiffs")


mm2gamma_zarr_v01 = (
    test_datasets / "MM20_zarr" / "mm2.0-20201209_4p_2t_5z_1c_512k_1.zarr"
)


hcs_ref = test_datasets / "20200812-CardiomyocyteDifferentiation14-Cycle1.zarr"


ndtiff_v2_datasets = subdirs(test_datasets, "MM20_pycromanager")


ndtiff_v2_ptcz = (
    test_datasets
    / "MM20_pycromanager"
    / "mm2.0-20210713_pm0.13.2_2p_3t_2c_7z_1"
)


ndtiff_v3_labeled_positions = test_datasets / "ndtiff_v3_labeled_positions"


@pytest.fixture
def csv_data_file_1(tmpdir):
    test_csv_1 = tmpdir / "well_names_1.csv"
    csv_data_1 = [
        ["B/03", "D/4"],
    ]
    with open(test_csv_1, mode="w", newline="") as csvfile:
        writer = csv.writer(csvfile)
        writer.writerows(csv_data_1)
    return test_csv_1


@pytest.fixture
def csv_data_file_2(tmpdir):
    test_csv_2 = tmpdir / "well_names_2.csv"
    csv_data_2 = [
        ["D/4", "B/03"],
    ]
    with open(test_csv_2, mode="w", newline="") as csvfile:
        writer = csv.writer(csvfile)
        writer.writerows(csv_data_2)
    return test_csv_2



"""OME-Zarr display settings (OMERO metadata)"""

import numpy as np
from PIL.ImageColor import colormap

from iohub.ngff.models import ChannelMeta, WindowDict

""" Dictionary with key works and most popular fluorescent probes """
CHANNEL_COLORS = {
    # emission around 510-525 nm
    "lime": [
        "Green",
        "GFP",
        "FITC",
        "mNeon",
    ],
    # emission around 580 - 610 nm
    "magenta": [
        "Magenta",
        "TXR",
        "RFP",
        "mScarlet",
        "mCherry",
    ],
    # emission around 440 - 460 nmm
    "blue": ["Blue", "DAPI", "BFP", "Hoechst"],
    "red": ["Red"],
    # Emission around 540-570 nm
    "yellow": ["Yellow", "Cy3"],
    # emission around 650-680 nm
    "orange": ["Orange", "Cy5", "Y5"],
}


def color_to_hex(color: str) -> str:
    """
    Convert the color string to HEX (i.e 'red' -> 'FF0000')
    (https://pillow.readthedocs.io/en/stable/_modules/PIL/ImageColor.html#getrgb)
    (https://www.w3.org/TR/css-color-3/#svg-color)

    Parameters
    ----------
    color : str The name of the color from the CSS3 Specifications

    Returns
    -------
    str the HEX value in uppercase and without the '#'
    """
    return colormap[color][1:].upper()


def channel_display_settings(
    chan_name: str,
    clim: tuple[float, float, float, float] | None = None,
    first_chan: bool = False,
):
    """This will create a dictionary used for OME-Zarr metadata.
    Allows custom contrast limits and channel names for display.
    Defaults everything to grayscale.

    Parameters
    ----------
    chan_name : str
        Desired name of the channel for display
    clim : tuple[float, float, float, float], optional
        Contrast limits (start, end, min, max)
    first_chan : bool, optional
        Whether or not this is the first channel of the dataset
        (display will be set to active),
        by default False

    Returns
    -------
    dict
        Display settings adherent to ome-zarr standards
    """
    U16_FMAX = float(np.iinfo(np.uint16).max)
    channel_settings = {
        "Retardance": (0.0, 100.0, 0.0, 1000.0),
        "Orientation": (0.0, np.pi, 0.0, np.pi),
        "Phase2D": (-0.2, 0.2, -10, 10),
        "Phase3D": (-0.2, 0.2, -10, 10),
        "BF": (0.0, 5.0, 0.0, U16_FMAX),
        "S0": (0.0, 1.0, 0.0, U16_FMAX),
        "S1": (-0.5, 0.5, -10.0, 10.0),
        "S2": (-0.5, 0.5, -10.0, 10.0),
        "S3": (-1.0, 1.0, -10.0, -10.0),
        "Other": (0, U16_FMAX, 0.0, U16_FMAX),
    }
    if clim is None:
        if chan_name in channel_settings.keys():
            clim = channel_settings[chan_name]
        else:
            clim = channel_settings["Other"]
    # Mapping channel name to color
    for key in CHANNEL_COLORS:
        # Does chan_name have any of the values in the CHANNEL_COLORS[key]
        # list as a substring?
        if any([value in chan_name for value in CHANNEL_COLORS[key]]):
            display_color = color_to_hex(key)
            break
        else:
            display_color = color_to_hex("white")

    window = WindowDict(start=clim[0], end=clim[1], min=clim[2], max=clim[3])
    return ChannelMeta(
        active=first_chan,
        coefficient=1.0,
        color=display_color,
        family="linear",
        inverted=False,
        label=chan_name,
        window=window,
    )



import re

import pytest
from xarray import DataArray

from iohub.mmstack import MMOmeTiffFOV, MMStack
from tests.conftest import (
    mm2gamma_ome_tiffs,
    mm2gamma_ome_tiffs_incomplete,
    mm1422_ome_tiffs,
)


def pytest_generate_tests(metafunc):
    if "ome_tiff" in metafunc.fixturenames:
        metafunc.parametrize("ome_tiff", mm2gamma_ome_tiffs + mm1422_ome_tiffs)


def test_mmstack_ctx(ome_tiff):
    with MMStack(ome_tiff) as mmstack:
        assert isinstance(mmstack, MMStack)
        assert len(mmstack) > 0
        assert "MMStack" in mmstack.__repr__()


def test_mmstack_nonexisting(tmpdir):
    with pytest.raises(FileNotFoundError):
        MMStack(tmpdir / "nonexisting")


def test_mmstack_getitem(ome_tiff):
    mmstack = MMStack(ome_tiff)
    assert isinstance(mmstack["0"], MMOmeTiffFOV)
    assert isinstance(mmstack[0], MMOmeTiffFOV)
    for key, fov in mmstack:
        assert isinstance(key, str)
        assert isinstance(fov, MMOmeTiffFOV)
        assert key in mmstack.__repr__()
        assert key in fov.__repr__()
    mmstack.close()


def test_mmstack_num_positions(ome_tiff):
    with MMStack(ome_tiff) as mmstack:
        p_match = re.search(r"_(\d+)p_", str(ome_tiff))
        assert len(mmstack) == int(p_match.group(1)) if p_match else 1


def test_mmstack_num_timepoints(ome_tiff):
    with MMStack(ome_tiff) as mmstack:
        t_match = re.search(r"_(\d+)t_", str(ome_tiff))
        for _, fov in mmstack:
            assert fov.shape[0] == int(t_match.group(1)) if t_match else 1


def test_mmstack_num_timepoints_incomplete():
    with MMStack(mm2gamma_ome_tiffs_incomplete) as mmstack:
        for name, fov in mmstack:
            assert fov.shape[0] == 20
            if int(name) >= 11:
                assert not fov[:].any()
            else:
                assert fov[:].any()


def test_mmstack_metadata(ome_tiff):
    with MMStack(ome_tiff) as mmstack:
        assert isinstance(mmstack.micromanager_metadata, dict)
        assert mmstack.micromanager_metadata["Summary"]
        assert mmstack.micromanager_summary
        if mmstack.stage_positions:
            assert "DefaultXYStage" in mmstack.stage_positions[0]


def test_fov_axes_names(ome_tiff):
    mmstack = MMStack(ome_tiff)
    for _, fov in mmstack:
        axes_names = fov.axes_names
        assert isinstance(axes_names, list)
        assert len(axes_names) == 5
        assert all([isinstance(name, str) for name in axes_names])
    mmstack.close()


def test_fov_getitem(ome_tiff):
    mmstack = MMStack(ome_tiff)
    for _, fov in mmstack:
        img: DataArray = fov[:]
        assert isinstance(img, DataArray)
        assert img.ndim == 5
        assert img[0, 0, 0, 0, 0] >= 0
        for ch in fov.channel_names:
            assert img.sel(T=0, Z=0, C=ch, Y=0, X=0) >= 0
    mmstack.close()


def test_fov_equal(ome_tiff):
    ref1 = MMStack(ome_tiff)
    ref2 = MMStack(ome_tiff)
    for (_, fov1), (_, fov2) in zip(ref1, ref2):
        assert fov1 == fov2
    ref1.close()
    ref2.close()


def test_fov_not_equal(ome_tiff):
    with MMStack(ome_tiff) as mmstack:
        if len(mmstack) < 2:
            # pass
            return
        assert mmstack["0"] != mmstack["1"]



import json
import logging
import os
from pathlib import Path
from typing import Literal

import numpy as np
import pytest
from ndtiff import Dataset
from tifffile import TiffFile

from iohub.convert import TIFFConverter
from iohub.ngff import Position, open_ome_zarr
from iohub.reader import MMStack, NDTiffDataset
from tests.conftest import (
    mm2gamma_ome_tiffs,
    ndtiff_v2_datasets,
    ndtiff_v3_labeled_positions,
)


def pytest_generate_tests(metafunc):
    if "mm2gamma_ome_tiff" in metafunc.fixturenames:
        metafunc.parametrize("mm2gamma_ome_tiff", mm2gamma_ome_tiffs)
    if "ndtiff_datasets" in metafunc.fixturenames:
        metafunc.parametrize(
            "ndtiff_datasets",
            ndtiff_v2_datasets + [ndtiff_v3_labeled_positions],
        )
    if "grid_layout" in metafunc.fixturenames:
        metafunc.parametrize("grid_layout", [True, False])
    if "chunks" in metafunc.fixturenames:
        metafunc.parametrize(
            "chunks", ["XY", "XYZ", (1, 1, 3, 256, 256), None]
        )


def _check_scale_transform(position: Position) -> None:
    """Check scale transformation of the highest resolution level."""
    tf = (
        position.metadata.multiscales[0]
        .datasets[0]
        .coordinate_transformations[0]
    )
    assert tf.type == "scale"
    assert tf.scale[:2] == [1.0, 1.0]


def _check_chunks(
    position: Position, chunks: Literal["XY", "XYZ"] | tuple[int] | None
) -> None:
    """Check chunk size of the highest resolution level."""
    img = position["0"]
    match chunks:
        case "XY":
            assert img.chunks == (1,) * 3 + img.shape[-2:]
        case "XYZ" | None:
            assert img.chunks == (1,) * 2 + img.shape[-3:]
        case tuple():
            assert img.chunks == chunks
        case _:
            assert False


def test_converter_inputs(mm2gamma_ome_tiff, tmpdir):
    # Test that output directory needs to end with ".zarr"
    output = tmpdir / "converted"
    with pytest.raises(ValueError):
        _ = TIFFConverter(mm2gamma_ome_tiff, output)


def test_converter_ometiff(mm2gamma_ome_tiff, grid_layout, chunks, tmpdir):
    logging.getLogger("tifffile").setLevel(logging.ERROR)
    output = tmpdir / "converted.zarr"
    converter = TIFFConverter(
        mm2gamma_ome_tiff, output, grid_layout=grid_layout, chunks=chunks
    )
    assert isinstance(converter.reader, MMStack)
    with TiffFile(next(mm2gamma_ome_tiff.glob("*.tif*"))) as tf:
        raw_array = tf.asarray()
        assert (
            converter.summary_metadata == tf.micromanager_metadata["Summary"]
        )
    assert np.prod([d for d in converter.dim if d > 0]) == np.prod(
        raw_array.shape
    )
    assert list(converter.metadata.keys()) == [
        "iohub_version",
        "Summary",
    ]
    converter()
    with open_ome_zarr(output, mode="r") as result:
        intensity = 0
        if grid_layout and converter.p > 1:
            assert len(result) < converter.p
        for pos_name, pos in result.positions():
            _check_scale_transform(pos)
            _check_chunks(pos, chunks)
            intensity += pos["0"][:].sum()
            assert os.path.isfile(
                os.path.join(
                    output, pos_name, "0", "image_plane_metadata.json"
                )
            )
        assert intensity == raw_array.sum()


@pytest.fixture(scope="module")
def example_ome_tiff() -> Path:
    for d in mm2gamma_ome_tiffs:
        if d.name == "mm2.0-20201209_4p_2t_5z_1c_512k_1":
            return d


@pytest.fixture(scope="function")
def mock_hcs_ome_tiff_reader(
    example_ome_tiff, monkeypatch: pytest.MonkeyPatch
):
    # dataset with 4 positions without HCS site names
    mock_stage_positions = [
        {"Label": "A1-Site_0"},
        {"Label": "A1-Site_1"},
        {"Label": "B4-Site_0"},
        {"Label": "H12-Site_0"},
    ]
    expected_ngff_name = {"A/1/0", "A/1/1", "B/4/0", "H/12/0"}
    monkeypatch.setattr(
        "iohub.convert.MMStack.stage_positions",
        mock_stage_positions,
    )
    return example_ome_tiff, expected_ngff_name


@pytest.fixture(scope="function")
def mock_non_hcs_ome_tiff_reader(
    example_ome_tiff, monkeypatch: pytest.MonkeyPatch
):
    # dataset with 4 positions without HCS site names
    mock_stage_positions = [
        {"Label": "0"},
        {"Label": "1"},
        {"Label": "2"},
        {"Label": "3"},
    ]
    monkeypatch.setattr(
        "iohub.convert.MMStack.stage_positions",
        mock_stage_positions,
    )
    return example_ome_tiff


def test_converter_ometiff_mock_hcs(mock_hcs_ome_tiff_reader, tmpdir):
    data, expected_ngff_name = mock_hcs_ome_tiff_reader
    output = tmpdir / "converted.zarr"
    converter = TIFFConverter(data, output, hcs_plate=True)
    converter()
    with open_ome_zarr(output, mode="r") as plate:
        assert expected_ngff_name == {name for name, _ in plate.positions()}


def test_converter_ometiff_mock_non_hcs(mock_non_hcs_ome_tiff_reader, tmpdir):
    data = mock_non_hcs_ome_tiff_reader
    output = tmpdir / "converted.zarr"
    with pytest.raises(ValueError, match="HCS position labels"):
        TIFFConverter(data, output, hcs_plate=True)


def test_converter_ometiff_hcs_numerical(example_ome_tiff, tmpdir):
    output = tmpdir / "converted.zarr"
    converter = TIFFConverter(example_ome_tiff, output, hcs_plate=True)
    converter()
    with open_ome_zarr(output, mode="r") as plate:
        for name, _ in plate.positions():
            for segment in name.split("/"):
                assert segment.isdigit()


def test_converter_ndtiff(ndtiff_datasets: Path, grid_layout, chunks, tmpdir):
    logging.getLogger("tifffile").setLevel(logging.ERROR)
    output = tmpdir / "converted.zarr"
    converter = TIFFConverter(
        ndtiff_datasets, output, grid_layout=grid_layout, chunks=chunks
    )
    assert isinstance(converter.reader, NDTiffDataset)
    raw_array = np.asarray(Dataset(str(ndtiff_datasets)).as_array())
    assert np.prod([d for d in converter.dim if d > 0]) == np.prod(
        raw_array.shape
    )
    assert list(converter.metadata.keys()) == [
        "iohub_version",
        "Summary",
    ]
    converter()
    with open_ome_zarr(output, mode="r") as result:
        intensity = 0
        for pos_name, pos in result.positions():
            _check_scale_transform(pos)
            _check_chunks(pos, chunks)
            intensity += pos["0"][:].sum()
            assert os.path.isfile(
                os.path.join(
                    output, pos_name, "0", "image_plane_metadata.json"
                )
            )
    assert intensity == raw_array.sum()
    with open(
        os.path.join(output, pos_name, "0", "image_plane_metadata.json")
    ) as f:
        metadata = json.load(f)
        key = "0/0/0"
        assert key in metadata
        assert "ElapsedTime-ms" in metadata[key]


def test_converter_ndtiff_v3_position_labels(tmpdir):
    output = tmpdir / "converted.zarr"
    converter = TIFFConverter(ndtiff_v3_labeled_positions, output)
    converter()
    with open_ome_zarr(output, mode="r") as result:
        assert result.channel_names == ["Channel0"]
        assert [name.split("/")[1] for name, _ in result.positions()] == [
            "Pos0",
            "Pos1",
            "Pos2",
        ]



import csv
from pathlib import Path

from iohub.ngff import open_ome_zarr


def rename_wells(zarr_store_path: str | Path, csv_file_path: str | Path):
    """
    Rename wells in a Zarr store based on a CSV file containing old and new
    well names.

    Parameters
    ----------
    zarr_store_path : str or Path
        Path to the Zarr store.
    csv_file_path : str or Path
        Path to the CSV file containing the old and new well names.

    Raises
    ------
    ValueError
        If a row in the CSV file does not have exactly two columns.
        If there is an error renaming a well in the Zarr store.

    Notes
    -----
    The CSV file should have two columns:
        - The first column contains the old well names.
        - The second column contains the new well names.

    Examples
    --------
    CSV file content:
    A/1,B/2
    A/2,B/2

    """

    # read and check csv
    name_pair_list = []
    with open(csv_file_path, mode="r", encoding="utf-8-sig") as csv_file:
        for row in csv.reader(csv_file):
            if len(row) != 2:
                raise ValueError(
                    f"Invalid row format: {row}."
                    f"Each row must have two columns."
                )
            name_pair_list.append([row[0].strip(), row[1].strip()])

    # rename each well while catching errors
    with open_ome_zarr(zarr_store_path, mode="a") as plate:
        for old, new in name_pair_list:
            print(f"Renaming {old} to {new}")
            try:
                plate.rename_well(old, new)
            except ValueError as e:
                print(f"Error renaming {old} to {new}: {e}")



import json
import logging
from importlib.metadata import version
from pathlib import Path
from typing import Literal

import numpy as np
from dask.array import to_zarr
from tqdm import tqdm
from tqdm.contrib.itertools import product
from tqdm.contrib.logging import logging_redirect_tqdm

from iohub.ngff.models import TransformationMeta
from iohub.ngff.nodes import Position, open_ome_zarr
from iohub.reader import MMStack, NDTiffDataset, read_images

__all__ = ["TIFFConverter"]
_logger = logging.getLogger(__name__)

MAX_CHUNK_SIZE = 500e6  # in bytes


def _create_grid_from_coordinates(
    xy_coords: list[tuple[float, float]], rows: int, columns: int
):
    """Create a grid from XY-position coordinates.

    Parameters
    ----------
    xy_coords : list[tuple[float, float]]
        (X, Y) stage position list in the order in which it was acquired.
    rows : int
        number of rows in the grid-like acquisition
    columns : int
        number of columns in the grid-like acquisition

    Returns
    -------
    NDArray
        A grid-like array mimicking the shape of the acquisition where the
        value in the array corresponds to the position index at that location.
    """

    coords = dict()
    coords_list = []
    for idx, pos in enumerate(xy_coords):
        coords[idx] = pos
        coords_list.append(pos)

    # sort by X and then by Y
    coords_list.sort(key=lambda x: x[0])
    coords_list.sort(key=lambda x: x[1])

    # reshape XY coordinates into their proper 2D shape
    grid = np.reshape(coords_list, (rows, columns, 2))
    pos_index_grid = np.zeros((rows, columns), "uint16")
    keys = list(coords.keys())
    vals = list(coords.values())

    for row in range(rows):
        for col in range(columns):
            # append position index (key) into a final grid
            # by indexed into the coordinate map (values)
            pos_index_grid[row, col] = keys[vals.index(list(grid[row, col]))]

    return pos_index_grid


class TIFFConverter:
    """Convert Micro-Manager TIFF formats
    (OME-TIFF, ND-TIFF) into HCS OME-Zarr.
    Each FOV will be written to a separate well in the plate layout.

    Parameters
    ----------
    input_dir : str | Path
        Input directory path
    output_dir : str | Path
        Output zarr directory path
    grid_layout : bool, optional
        Whether to lay out the positions in a grid-like format
        based on how the data was acquired
        (useful for tiled acquisitions), by default False
    chunks : tuple[int] or Literal['XY', 'XYZ'], optional
        Chunk size of the output Zarr arrays, by default None
        (chunk by XYZ volumes or 500 MB size limit, whichever is smaller)
    hcs_plate : bool, optional
        Create NGFF HCS layout based on position names from the
        HCS Site Generator in Micro-Manager (only available for OME-TIFF),
        and is ignored for other formats, by default None
        (attempt to apply to OME-TIFF datasets, disable this with ``False``)

    Notes
    -----
    The image plane metadata for each FOV is aggregated into a JSON file,
    and placed under the Zarr array directory
    (e.g. ``/row/column/fov/0/image_plane_metadata.json``).
    """

    def __init__(
        self,
        input_dir: str | Path,
        output_dir: str | Path,
        grid_layout: int = False,
        chunks: tuple[int] | Literal["XY", "XYZ"] = None,
        hcs_plate: bool = None,
    ):
        _logger.debug("Checking output.")
        output_dir = Path(output_dir)
        if ".zarr" not in output_dir.suffixes:
            raise ValueError("Please specify .zarr at the end of your output")
        self.output_dir = output_dir
        _logger.info("Initializing data.")
        self.reader = read_images(input_dir)
        if reader_type := type(self.reader) not in (
            MMStack,
            NDTiffDataset,
        ):
            raise TypeError(
                f"Reader type {reader_type} not supported for conversion."
            )
        _logger.debug("Finished initializing data.")
        self.summary_metadata = self.reader.micromanager_summary
        self.save_name = output_dir.name
        _logger.debug("Getting dataset summary information.")
        self.coord_map = dict()
        self.p = len(self.reader)
        self.t = self.reader.frames
        self.c = self.reader.channels
        self.z = self.reader.slices
        self.y = self.reader.height
        self.x = self.reader.width
        self.dim = (self.p, self.t, self.c, self.z, self.y, self.x)
        self.prefix_list = []
        self.hcs_plate = hcs_plate
        self._check_hcs_sites()
        self._get_pos_names()
        _logger.info(
            f"Found Dataset {input_dir} with "
            f"dimensions (P, T, C, Z, Y, X): {self.dim}"
        )
        self.metadata = dict()
        self.metadata["iohub_version"] = version("iohub")
        self.metadata["Summary"] = self.summary_metadata
        if grid_layout:
            if hcs_plate:
                raise ValueError(
                    "grid_layout and hcs_plate must not be both true"
                )
            _logger.info("Generating HCS plate level grid.")
            try:
                self.position_grid = _create_grid_from_coordinates(
                    *self._get_position_coords()
                )
            except ValueError as e:
                _logger.warning(f"Failed to generate grid layout: {e}")
                self._make_default_grid()
        else:
            self._make_default_grid()
        self.chunks = self._gen_chunks(chunks)
        self.transform = self._scale_voxels()

    def _check_hcs_sites(self):
        if self.hcs_plate:
            self.hcs_sites = self.reader.hcs_position_labels
        elif self.hcs_plate is None:
            try:
                self.hcs_sites = self.reader.hcs_position_labels
                self.hcs_plate = True
            except ValueError:
                _logger.debug(
                    "HCS sites not detected, "
                    "dumping all position into a single row."
                )

    def _make_default_grid(self):
        if isinstance(self.reader, NDTiffDataset):
            self.position_grid = np.array([self.pos_names])
        else:
            self.position_grid = np.expand_dims(
                np.arange(self.p, dtype=int), axis=0
            )

    def _get_position_coords(self):
        """Get the position coordinates from the reader metadata.

        Raises:
            ValueError: If stage positions are not available.

        Returns:
            list: XY stage position coordinates.
            int: Number of grid rows.
            int: Number of grid columns.
        """
        rows = set()
        cols = set()
        xy_coords = []

        # TODO: account for non MM2gamma meta?
        if not self.reader.stage_positions:
            raise ValueError("Stage positions not available.")
        for idx, pos in enumerate(self.reader.stage_positions):
            try:
                xy_stage = pos["DefaultXYStage"]
                stage_pos = pos[xy_stage]
            except KeyError:
                raise ValueError(
                    f"Stage position is not available for position {idx}"
                )
            xy_coords.append(stage_pos)
            try:
                rows.add(pos["GridRow"])
                cols.add(pos["GridCol"])
            except KeyError:
                raise ValueError(
                    f"Grid indices not available for position {idx}"
                )

        return xy_coords, len(rows), len(cols)

    def _get_pos_names(self):
        """Append a list of pos names in ascending order
        (order in which they were acquired).
        """
        self.pos_names = []
        for p in range(self.p):
            try:
                name = self.reader.stage_positions[p]["Label"]
            except (IndexError, KeyError):
                name = str(p)
            self.pos_names.append(name)

    def _gen_chunks(self, input_chunks):
        if not input_chunks:
            _logger.debug("No chunk size specified, using ZYX.")
            chunks = [1, 1, self.z, self.y, self.x]
        elif isinstance(input_chunks, tuple):
            chunks = list(input_chunks)
        elif isinstance(input_chunks, str):
            if input_chunks.lower() == "xy":
                chunks = [1, 1, 1, self.y, self.x]
            elif input_chunks.lower() == "xyz":
                chunks = [1, 1, self.z, self.y, self.x]
            else:
                raise ValueError(f"{input_chunks} chunks are not supported.")
        else:
            raise TypeError(
                f"Chunk type {type(input_chunks)} is not supported."
            )

        # limit chunks to MAX_CHUNK_SIZE bytes
        bytes_per_pixel = np.dtype(self.reader.dtype).itemsize
        # it's OK if a single image is larger than MAX_CHUNK_SIZE
        while (
            chunks[-3] > 1
            and np.prod(chunks, dtype=np.int64) * bytes_per_pixel
            > MAX_CHUNK_SIZE
        ):
            chunks[-3] = np.ceil(chunks[-3] / 2).astype(int)

        _logger.debug(f"Zarr store chunk size will be set to {chunks}.")

        return tuple(chunks)

    def _scale_voxels(self):
        return [
            TransformationMeta(
                type="scale", scale=[1.0, 1.0, *self.reader.zyx_scale]
            )
        ]

    def _init_zarr_arrays(self):
        self.writer = open_ome_zarr(
            self.output_dir,
            layout="hcs",
            mode="w-",
            channel_names=self.reader.channel_names,
            version="0.4",
        )
        self.zarr_position_names = []
        arr_kwargs = {
            "name": "0",
            "shape": (
                self.t if self.t != 0 else 1,
                self.c if self.c != 0 else 1,
                self.z if self.z != 0 else 1,
                self.y,
                self.x,
            ),
            "dtype": self.reader.dtype,
            "chunks": self.chunks,
            "transform": self.transform,
        }
        if self.hcs_plate:
            self._init_hcs_arrays(arr_kwargs)
        else:
            self._init_grid_arrays(arr_kwargs)

    def _init_hcs_arrays(self, arr_kwargs):
        for row, col, fov in self.hcs_sites:
            self._create_zeros_array(row, col, fov, arr_kwargs)
        _logger.info(
            "Created HCS NGFF layout from Micro-Manager HCS position labels."
        )
        self.writer.print_tree()

    def _init_grid_arrays(self, arr_kwargs):
        for row, columns in enumerate(self.position_grid):
            for column in columns:
                self._create_zeros_array(row, column, "0", arr_kwargs)

    def _create_zeros_array(
        self, row_name: str, col_name: str, pos_name: str, arr_kwargs: dict
    ) -> Position:
        pos = self.writer.create_position(row_name, col_name, pos_name)
        self.zarr_position_names.append(pos.zgroup.name)
        _ = pos.create_zeros(**arr_kwargs)
        pos.metadata.omero.name = self.pos_names[
            len(self.zarr_position_names) - 1
        ]
        pos.dump_meta()

    def _convert_image_plane_metadata(self, fov, zarr_name: str):
        position_image_plane_metadata = {}
        sorted_keys = []
        for t_idx, c_idx in product(
            range(self.t),
            range(self.c),
            desc="Converting frame metadata",
            unit="frame",
            leave=False,
            ncols=80,
        ):
            c_key = c_idx
            if isinstance(self.reader, NDTiffDataset):
                c_key = (
                    self.reader.channel_names[c_idx]
                    if self.reader.str_channel_axis
                    else c_idx
                )
            missing_data_warning_issued = False
            for z_idx in range(self.z):
                metadata = fov.frame_metadata(t=t_idx, c=c_key, z=z_idx)
                if metadata is None:
                    if not missing_data_warning_issued:
                        missing_data_warning_issued = True
                        _logger.warning(
                            f"Cannot load data at P: {zarr_name}, T: {t_idx}, "
                            f"C: {c_idx}, filling with zeros. Raw data may be "
                            "incomplete."
                        )
                    continue
                if not sorted_keys:
                    # Sort keys, ordering keys without dashes first
                    sorted_keys = sorted(
                        metadata.keys(), key=lambda x: ("-" in x, x)
                    )

                sorted_metadata = {key: metadata[key] for key in sorted_keys}
                # T/C/Z
                frame_key = "/".join([str(i) for i in (t_idx, c_idx, z_idx)])
                position_image_plane_metadata[frame_key] = sorted_metadata
        with open(
            self.output_dir / zarr_name / "image_plane_metadata.json",
            mode="x",
        ) as metadata_file:
            json.dump(position_image_plane_metadata, metadata_file, indent=4)

    def __call__(self) -> None:
        """
        Runs the conversion.

        Examples
        --------
        >>> from iohub.convert import TIFFConverter
        >>> converter = TIFFConverter("input/path/", "output/path/")
        >>> converter()
        """
        _logger.debug("Setting up Zarr store.")
        self._init_zarr_arrays()
        # Run through every coordinate and convert in acquisition order
        _logger.debug("Converting images.")
        with logging_redirect_tqdm():
            for zarr_pos_name, (_, fov) in tqdm(
                zip(self.zarr_position_names, self.reader, strict=True),
                total=len(self.reader),
                desc="Converting images",
                unit="FOV",
                ncols=80,
            ):
                zarr_img = self.writer[zarr_pos_name]["0"]
                to_zarr(fov.xdata.data.rechunk(self.chunks), zarr_img)
                self._convert_image_plane_metadata(fov, zarr_img.path)
        self.writer.zgroup.attrs.update(self.metadata)
        self.writer.close()
        self.reader.close()



from pathlib import Path
from typing import Callable, List

import click
from natsort import natsorted

from iohub.ngff import Plate, open_ome_zarr


def _validate_and_process_paths(
    ctx: click.Context, opt: click.Option, value: List[str]
) -> list[Path]:
    # Sort and validate the input paths,
    # expanding plates into lists of positions
    input_paths = [Path(path) for path in natsorted(value)]
    for path in input_paths:
        with open_ome_zarr(path, mode="r") as dataset:
            if isinstance(dataset, Plate):
                plate_path = input_paths.pop()
                for position in dataset.positions():
                    input_paths.append(plate_path / position[0])

    return input_paths


def input_position_dirpaths() -> Callable:
    def decorator(f: Callable) -> Callable:
        return click.option(
            "--input-position-dirpaths",
            "-i",
            cls=OptionEatAll,
            type=tuple,
            required=True,
            callback=_validate_and_process_paths,
            help=(
                "List of paths to input positions, "
                "each with the same TCZYX shape. "
                "Supports wildcards e.g. 'input.zarr/*/*/*'."
            ),
        )(f)

    return decorator


# Copied directly from https://stackoverflow.com/a/48394004
# Enables `-i ./input.zarr/*/*/*`
class OptionEatAll(click.Option):
    def __init__(self, *args, **kwargs):
        self.save_other_options = kwargs.pop("save_other_options", True)
        nargs = kwargs.pop("nargs", -1)
        assert nargs == -1, "nargs, if set, must be -1 not {}".format(nargs)
        super(OptionEatAll, self).__init__(*args, **kwargs)
        self._previous_parser_process = None
        self._eat_all_parser = None

    def add_to_parser(self, parser, ctx):
        def parser_process(value, state):
            # method to hook to the parser.process
            done = False
            value = [value]
            if self.save_other_options:
                # grab everything up to the next option
                while state.rargs and not done:
                    for prefix in self._eat_all_parser.prefixes:
                        if state.rargs[0].startswith(prefix):
                            done = True
                    if not done:
                        value.append(state.rargs.pop(0))
            else:
                # grab everything remaining
                value += state.rargs
                state.rargs[:] = []
            value = tuple(value)

            # call the actual process
            self._previous_parser_process(value, state)

        retval = super(OptionEatAll, self).add_to_parser(parser, ctx)
        for name in self.opts:
            our_parser = parser._long_opt.get(name) or parser._short_opt.get(
                name
            )
            if our_parser:
                self._eat_all_parser = our_parser
                self._previous_parser_process = our_parser.process
                our_parser.process = parser_process
                break
        return retval



import pathlib

import click

from iohub import open_ome_zarr
from iohub._version import __version__
from iohub.cli.parsing import input_position_dirpaths
from iohub.convert import TIFFConverter
from iohub.reader import print_info
from iohub.rename_wells import rename_wells

VERSION = __version__

_DATASET_PATH = click.Path(
    exists=True, file_okay=False, resolve_path=True, path_type=pathlib.Path
)


@click.group()
@click.help_option("-h", "--help")
@click.version_option(version=VERSION)
def cli():
    """\u001b[34;1m iohub: N-dimensional bioimaging I/O \u001b[0m"""


@cli.command()
@click.help_option("-h", "--help")
@click.argument(
    "files",
    nargs=-1,
    required=True,
    type=_DATASET_PATH,
)
@click.option(
    "--verbose",
    "-v",
    is_flag=True,
    help="Show usage guide to open dataset in Python "
    "and full tree for HCS Plates in OME-Zarr",
)
def info(files, verbose):
    """View basic metadata of a list of FILES.

    Supported formats are Micro-Manager-acquired TIFF datasets
    (single-page TIFF, multi-page OME-TIFF, NDTIFF)
    and OME-Zarr (v0.1 linear HCS layout and all v0.4 layouts).
    """
    for file in files:
        click.echo(f"Reading file:\t {file}")
        print_info(file, verbose=verbose)


@cli.command()
@click.help_option("-h", "--help")
@click.option(
    "--input",
    "-i",
    required=True,
    type=_DATASET_PATH,
    help="Input Micro-Manager TIFF dataset directory",
)
@click.option(
    "--output",
    "-o",
    required=True,
    type=click.Path(exists=False, resolve_path=True),
    help="Output zarr store (/**/converted.zarr)",
)
@click.option(
    "--grid-layout",
    "-g",
    required=False,
    is_flag=True,
    help="Arrange FOVs in a row/column grid layout for tiled acquisition",
)
@click.option(
    "--chunks",
    "-c",
    required=False,
    default="XYZ",
    help="Zarr chunk size given as 'XY', 'XYZ', or a tuple of chunk "
    "dimensions. If 'XYZ', chunk size will be limited to 500 MB.",
)
def convert(input, output, grid_layout, chunks):
    """Converts Micro-Manager TIFF datasets to OME-Zarr"""
    converter = TIFFConverter(
        input_dir=input,
        output_dir=output,
        grid_layout=grid_layout,
        chunks=chunks,
    )
    converter()


@cli.command()
@click.help_option("-h", "--help")
@input_position_dirpaths()
@click.option(
    "--t-scale",
    "-t",
    required=False,
    type=float,
    help="New t scale",
)
@click.option(
    "--z-scale",
    "-z",
    required=False,
    type=float,
    help="New z scale",
)
@click.option(
    "--y-scale",
    "-y",
    required=False,
    type=float,
    help="New y scale",
)
@click.option(
    "--x-scale",
    "-x",
    required=False,
    type=float,
    help="New x scale",
)
@click.option(
    "--image",
    required=False,
    help="Image name to set scale for. Default is '0'",
)
def set_scale(
    input_position_dirpaths,
    t_scale=None,
    z_scale=None,
    y_scale=None,
    x_scale=None,
    image=None,
):
    """Update scale metadata in OME-Zarr datasets.

    >> iohub set-scale -i input.zarr/*/*/* -t 1.0 -z 1.0 -y 0.5 -x 0.5

    Supports setting a single axis at a time:

    >> iohub set-scale -i input.zarr/*/*/* -z 2.0
    """
    if image is None:
        image = "0"
    for input_position_dirpath in input_position_dirpaths:
        with open_ome_zarr(
            input_position_dirpath, layout="fov", mode="r+"
        ) as dataset:
            for name, value in zip(
                ["t", "z", "y", "x"], [t_scale, z_scale, y_scale, x_scale]
            ):
                if value is None:
                    continue
                dataset.set_scale(image, name, value)


@cli.command(name="rename-wells")
@click.help_option("-h", "--help")
@click.option(
    "-i",
    "--input",
    "zarrfile",
    type=click.Path(exists=True, file_okay=True, dir_okay=True),
    required=True,
    help="Path to the input Zarr file.",
)
@click.option(
    "-c",
    "--csv",
    "csvfile",
    type=click.Path(exists=True, file_okay=True, dir_okay=False),
    required=True,
    help="Path to the CSV file containing old and new well names.",
)
def rename_wells_command(zarrfile, csvfile):
    """Rename wells in an plate.

    >> iohub rename-wells -i plate.zarr -c names.csv

    The CSV file must have two columns with old and new names in the form:
    ```
    A/1,B/2
    A/2,B/2
    ```
    """
    rename_wells(zarrfile, csvfile)



"""
Single-FOV OME-Zarr
===================

This script shows how to create a single-FOV, single-scale OME-Zarr dataset,
read data in read-only mode,
append an extra time point to an existing dataset,
and adding a new channel to an existing dataset.
"""

# %%
import os
from tempfile import TemporaryDirectory

import numpy as np

from iohub.ngff import open_ome_zarr

# %%
# Set storage path
tmp_dir = TemporaryDirectory()
store_path = os.path.join(tmp_dir.name, "ome.zarr")
print("Zarr store path", store_path)

# %%
# Write 5D data to a new Zarr store

tczyx = np.random.randint(
    0, np.iinfo(np.uint16).max, size=(5, 2, 3, 32, 32), dtype=np.uint16
)

with open_ome_zarr(
    store_path, layout="fov", mode="a", channel_names=["DAPI", "GFP"]
) as dataset:
    dataset["img"] = tczyx

# %%
# Opening in read-only mode prevents writing

with open_ome_zarr(store_path, layout="auto", mode="r") as dataset:
    img = dataset["img"]
    print(img)
    print(img.numpy())
    try:
        img[0, 0, 0, 0, 0] = 0
    except Exception as e:
        print(f"Writing was rejected: {e}")

# %%
# Append a new timepoint to an existing dataset

new_1czyx = np.random.randint(
    0, np.iinfo(np.uint16).max, size=(1, 2, 3, 32, 32), dtype=np.uint16
)

with open_ome_zarr(store_path, layout="fov", mode="r+") as dataset:
    img = dataset["img"]
    print(img.shape)
    img.append(new_1czyx, axis=0)
    print(img.shape)

# %%
# Modify channels

# Open the dataset used above
dataset = open_ome_zarr(store_path, mode="r+")
dataset.print_tree()

# Append a new channel and write a Z-stack
new_zyx = np.random.randint(
    0, np.iinfo(np.uint16).max, size=(3, 32, 32), dtype=np.uint16
)
dataset.append_channel("New", resize_arrays=True)
dataset["img"][0, 2] = new_zyx
print(dataset.channel_names)
dataset.print_tree()

# Rename the new channel
dataset.rename_channel("New", "Renamed")
print(dataset.channel_names)

# Write new data to the channel
new_tzyx = np.random.randint(
    0, np.iinfo(np.uint16).max, size=(6, 3, 32, 32), dtype=np.uint16
)
c_idx = dataset.get_channel_index("Renamed")
dataset["img"][:, c_idx] = new_tzyx

# Which is equivalent to:
if False:  # remove this line
    dataset.update_channel("Renamed", target="img", data=new_tzyx)

# Close the dataset
dataset.close()

# %%
# Try viewing the images with napari-ome-zarr

# %%
# Clean up
tmp_dir.cleanup()


import glob
import os
import re

import numpy as np
import tifffile as tiff
import zarr

from iohub._deprecated.reader_base import ReaderBase


class UPTIReader(ReaderBase):
    """
    Reader for UPTI raw data.
    Accepts both new live UPTI and older UPTI format.
    """

    def __init__(self, folder: str, extract_data: bool = False):
        # check if folder exists
        if not os.path.isdir(folder):
            raise ValueError("folder does not exist")

        # Initializate data parameters
        self.data_folder = folder
        if glob.glob(os.path.join(folder, "*.tif")) == []:
            self.files = glob.glob(os.path.join(folder, "*.tiff"))
        else:
            self.files = glob.glob(os.path.join(folder, "*.tif"))
        info_img = tiff.imread(self.files[0])
        self.dtype = info_img.dtype
        (self.height, self.width) = info_img.shape
        self.positions = 1
        self.frames = 1
        self.patterns = 0
        self.states = 0
        self.slices = 0
        self.channel_names = []

        # map files and update data parameters
        self._map_files()
        self.channels = len(self.channel_names)
        self.z_step_size = None

        # initialize metadata
        self.position_arrays = dict()

        if extract_data:
            for i in range(self.positions):
                self._create_position_array(i)

    def _map_files(self):
        """
        Creates a map of (pattern, state, slice) coordinate
        with corresponding file.
        Uses regex parsing and assumes a consistent file naming structure.

        Returns
        -------

        """

        self.file_map = dict()

        states = True if "State" in self.files[0] else False

        # Loop through the files and use regex to grab patterns and states,
        # **hardcoded file name structure**
        for file in self.files:
            pattern = re.search(r"pattern_\d\d\d", file).group(0)
            pattern_idx = int(pattern.strip("pattern_"))
            if re.search(r"z_\d\d\d", file) is None:
                z = 0
            else:
                z = int(re.search(r"z_\d\d\d", file).group(0).strip("z_"))
            state_idx = 0

            # update dimensionality as we read the file names
            if pattern_idx + 1 > self.patterns:
                self.patterns = pattern_idx + 1

            if z + 1 > self.slices:
                self.slices = z + 1

            # if live upti, grab the states, otherwise just use patterns
            if states:
                state = re.search(r"State_\d\d\d", file).group(0)
                state_idx = int(state.strip("State_"))

                if f"{pattern}_{state}" not in self.channel_names:
                    self.channel_names.append(f"{pattern}_{state}")
            else:
                if pattern not in self.channel_names:
                    self.channel_names.append(pattern)

            self.file_map[(pattern_idx, state_idx, z)] = file

            # sorts list by pattern 0 --> state_0, state_1, state_2, state_3
            self.channel_names.sort(
                key=lambda x: re.sub(r"pattern_\d\d\d", "", x)
            )
            self.channel_names.sort(
                key=lambda x: re.sub(r"State_\d\d\d", "", x)
            )

    def _create_position_array(self, pos):
        """
        maps all of the tiff data into a virtual zarr store in memory
        for a given position

        Parameters
        ----------
        pos:            (int) index of the position to create array under

        Returns
        -------

        """

        self.position_arrays[pos] = zarr.zeros(
            shape=(
                self.frames,
                self.channels,
                self.slices,
                self.height,
                self.width,
            ),
            chunks=(1, 1, 1, self.height, self.width),
            dtype=self.dtype,
        )
        # add all the images with this specific dimension.
        # Will be blank images if dataset
        # is incomplete
        for t in range(self.frames):
            for c in range(self.channels):
                for z in range(self.slices):
                    self.position_arrays[pos][t, c, z] = self._get_image(c, z)

    def _get_image(self, c, z):
        """
        Gets image at specific channel, z coordinate.
        Makes sure that the channel index
        specified corresponds to the channel_name index.

        Parameters
        ----------
        c:              (int) channel index.  Maps to channel_name index
        z:              (int) z/slice index.

        Returns
        -------
        img             (nd-array) numpy array of dimensions (Y, X)
        """

        chan_name = self.channel_names[c]
        pattern = int(
            re.search(r"pattern_\d\d\d", chan_name).group(0).strip("pattern_")
        )
        state = re.search(r"State_\d\d\d", chan_name)
        state_idx = 0 if not state else int(state.group(0).strip("State_"))

        fn = self.file_map[(pattern, state_idx, z)]
        img = zarr.open(tiff.imread(fn, aszarr=True), "r")

        return img

    def get_zarr(self, position):
        """
        return a zarr array for a given position.  Allows for only one position

        Parameters
        ----------
        position:       (int) position

        Returns
        -------
        position:       (zarr.array)

        """
        if position > self.positions - 1:
            raise ValueError(
                "Entered position is greater than "
                "the number of positions in the data"
            )

        if position not in self.position_arrays.keys():
            self._create_position_array(position)
        return self.position_arrays[position]

    def get_array(self, position):
        """
        return a numpy array for a given position.
        Allows for only one position

        Parameters
        ----------
        position:   (int) position

        Returns
        -------
        position:   (np.ndarray)

        """

        if position > self.positions - 1:
            raise ValueError(
                "Entered position is greater than "
                "the number of positions in the data"
            )

        # if position hasn't been initialized in memory, do that.
        if position not in self.position_arrays.keys():
            self._create_position_array(position)

        return np.array(self.position_arrays[position])

    def get_num_positions(self) -> int:
        return self.positions

    @property
    def shape(self):
        return self.frames, self.channels, self.slices, self.height, self.width



from __future__ import annotations

import re
from pathlib import Path

from xarray import DataArray

from iohub.fov import BaseFOV, BaseFOVMapping


class MicroManagerFOV(BaseFOV):
    def __init__(self, parent: MicroManagerFOVMapping, key: int) -> None:
        self._position = key
        self._parent = parent

    def __repr__(self) -> str:
        return (
            f"Type: {type(self)}\n"
            f"Parent: {self.parent}\n"
            f"FOV key: {self._position}\n"
            f"Data:\n"
        ) + self.xdata.__repr__()

    def __eq__(self, other: BaseFOV) -> bool:
        if not isinstance(other, type(self)):
            return False
        return (self._position == other._position) and (
            self.root.absolute() == other.root.absolute()
        )

    @property
    def parent(self) -> MicroManagerFOVMapping:
        return self._parent

    @property
    def root(self) -> Path:
        return self.parent.root

    @property
    def zyx_scale(self) -> tuple[float, float, float]:
        return self.parent.zyx_scale

    @property
    def channel_names(self) -> list[str]:
        return self.parent.channel_names

    @property
    def xdata(self) -> DataArray:
        raise NotImplementedError

    def frame_metadata(self, t: int, z: int, c: int) -> dict | None:
        """
        Return image plane metadata for a given camera frame.

        Parameters
        ----------
        t : int
            Time index.
        z : int
            Z slice index.
        c : int
            Channel index.

        Returns
        -------
        dict | None
            Image plane metadata. None if not available.
        """
        raise NotImplementedError


class MicroManagerFOVMapping(BaseFOVMapping):
    def __init__(self):
        self._root: Path = None
        self._mm_meta: dict = None
        self._stage_positions: list[dict[str, str | float]] = []
        self.channel_names: list[str] = None

    def __repr__(self) -> str:
        return (f"Type: {type(self)}\nData:\n") + self.xdata.__repr__()

    @property
    def root(self) -> Path:
        """Root directory of the dataset."""
        return self._root

    @property
    def micromanager_metadata(self) -> dict | None:
        return self._mm_meta

    @micromanager_metadata.setter
    def micromanager_metadata(self, value):
        if not isinstance(value, dict):
            raise TypeError(
                f"Type of `mm_meta` should be `dict`, got `{type(value)}`."
            )
        self._mm_meta = value

    @property
    def micromanager_summary(self) -> dict | None:
        """Micro-manager summary metadata."""
        return self._mm_meta.get("Summary", None)

    @property
    def stage_positions(self):
        return self._stage_positions

    @stage_positions.setter
    def stage_positions(self, value):
        if not isinstance(value, list):
            raise TypeError(
                "Type of `stage_position` should be `list`, "
                f"got `{type(value)}`."
            )
        self._stage_positions = value

    @property
    def hcs_position_labels(self):
        """Parse plate position labels generated by the HCS position generator,
        e.g. 'A1-Site_0', '1-Pos000_000', or 'Pos-9-000_000', and split into
        row, column, and FOV names.

        Returns
        -------
        list[tuple[str, str, str]]
            FOV name paths, e.g. ('A', '1', '0'), ('0', '1', '000000'), or
            ('0', '9', '000000').
        """
        if not self.stage_positions:
            raise ValueError("Stage position metadata not available.")

        try:
            labels = [pos["Label"] for pos in self.stage_positions]
        except KeyError:
            raise ValueError("Stage positions do not have labels.")

        # See https://chatgpt.com/share/e/670097cc-2854-8008-bd33-b54cad7c99b9
        pattern = re.compile(
            r"([A-Z])(\d+)-Site_(\d+)|"
            r"Pos-(\d+)-(\d+)_(\d+)|"
            r"(\d+)-Pos(\d+)_(\d+)"
        )
        row_col_fov = []
        for label in labels:
            if (match := re.match(pattern, label)) is not None:
                if match.group(1):  # "A1-Site_0" case
                    row_col_fov.append(
                        (match.group(1), match.group(2), match.group(3))
                    )
                elif match.group(4):  # "Pos-5-000_005" case
                    row_col_fov.append(
                        ("0", match.group(4), match.group(5) + match.group(6))
                    )
                elif match.group(7):  # "1-Pos000_000" case
                    row_col_fov.append(
                        ("0", match.group(7), match.group(8) + match.group(9))
                    )

        if not row_col_fov:
            raise ValueError(
                "HCS position labels are in the format of "
                "'A1-Site_0', 'H12-Site_1', or '1-Pos000_000', "
                f"or 'Pos-1-000_000'. Got labels {labels}"
            )

        return row_col_fov

    @property
    def zyx_scale(self) -> tuple[float, float, float]:
        """ZXY pixel size in micrometers."""
        raise NotImplementedError
